package aexp.stabilizedcompass;

import android.app.Service;
import android.content.Intent;
import android.content.Context;
import android.content.SharedPreferences;
import android.content.IntentFilter;
import android.content.BroadcastReceiver;
import android.hardware.Sensor;
import android.hardware.SensorEvent;
import android.hardware.SensorEventListener;
import android.os.Environment;
import android.os.PowerManager;
import android.os.IBinder;
import android.os.RemoteException;
import android.os.DeadObjectException;
import android.hardware.SensorManager;
import android.util.Log;
import android.util.LogPrinter;
import java.io.PrintWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.List;
import java.util.Date;

public class SamplingService extends Service implements SensorEventListener {
	static final String LOG_TAG = "COMPASSACCEL";
	static final boolean DEBUG = true;
	static final int SAMPLECTR_MOD = 100;
	static final double CALIBRATING_ACCEL_LIMIT = 0.1;
	static final double NOMOTION_ACCEL_LIMIT = 0.1;
// Maximum time between compass and accel samples to belong together 
	static final long MAX_ACCEL_COMPASS_DIFF = 60L * 1000L * 1000L; 
	static final long DIFF_UPDATE_TIMEOUT = 100L;
	public static final int  ENGINESTATES_IDLE = 0;
	public static final int ENGINESTATES_CALIBRATING = 1;
	public static final int ENGINESTATES_MEASURING = 2;
	public static final int ENGINESTATES_TRANSIENT = 100;
	public static final int ENGINESTATES_XGONE = 100;
	public static final int ENGINESTATES_YGONE = 101;
	public static final int ENGINESTATES_ZGONE = 102;
	
	public static final int SENSORTYPE_NA = 0;
	public static final int SENSORTYPE_ACCEL = 1;
	public static final int SENSORTYPE_COMPASS = 0;
	public static final int IDX_X = 0;
	public static final int IDX_Y = 1;
	public static final int IDX_Z = 2;
		
	public int onStartCommand(Intent intent, int flags, int startId) {
		super.onStartCommand( intent, flags, startId );
		Log.d( LOG_TAG, "onStartCommand" );
		stopSampling();		// just in case the activity-level service management fails
        rate = SensorManager.SENSOR_DELAY_FASTEST;
		sensorManager = (SensorManager)getSystemService( SENSOR_SERVICE  );
		startSampling();
		Log.d( LOG_TAG, "onStartCommand ends" );
		return START_NOT_STICKY;
	}

	public void onDestroy() {
		super.onDestroy();
		Log.d( LOG_TAG, "onDestroy" );
		stopSampling();
	}

	public IBinder onBind(Intent intent) {
		return serviceBinder;	// cannot bind
	}

// SensorEventListener
    public void onAccuracyChanged (Sensor sensor, int accuracy) {
    }

    public void onSensorChanged(SensorEvent sensorEvent) {
		processSample( sensorEvent );
    }


	private void stopSampling() {
		if( !samplingStarted )
			return;
        if( sensorManager != null ) {
			Log.d( LOG_TAG, "unregisterListener/SamplingService" );
            sensorManager.unregisterListener( this );
		}
        if( captureFile != null ) {
            captureFile.close();
			captureFile = null;
        }
		samplingStarted = false;
		setState( ENGINESTATES_IDLE);
	}

	private void startSampling() {
		if( samplingStarted )
			return;
       	List<Sensor> sensors = sensorManager.getSensorList( Sensor.TYPE_ACCELEROMETER  );
      	accelSensor = sensors.size() == 0 ? null : sensors.get( 0 );
       	sensors = sensorManager.getSensorList( Sensor.TYPE_MAGNETIC_FIELD );
      	compassSensor = sensors.size() == 0 ? null : sensors.get( 0 );
      	initSampling();
       	
      	if( ( accelSensor != null ) && ( compassSensor != null ) ) {
			Log.d( LOG_TAG, "registerListener/SamplingService" );
           	sensorManager.registerListener( 
                            this, 
                            accelSensor,
                            rate );
           	sensorManager.registerListener( 
                    this, 
                    compassSensor,
                    rate );
		} else {
			Log.d( LOG_TAG, 
				"Sensor(s) missing: accelSensor: "+
				accelSensor+
				"; compassSensor: "+
				compassSensor);
		}
      	captureFile = null;
      	if( DEBUG ) {
      		File captureFileName = new File( 
      				Environment.getExternalStorageDirectory(), 
      				"capture.csv" );
      		try {
      			captureFile = new PrintWriter( new FileWriter( captureFileName, false ) );
      		} catch( IOException ex ) {
      			Log.e( LOG_TAG, ex.getMessage(), ex );
      		}
      	}
      	samplingStarted = true;
	}

	private void initSampling() {
		sampleCounter = 0;
	    compassXMinSet = false;
	    compassXMaxSet = false;
	    compassYMinSet = false;
	    compassYMaxSet = false;
	    compassZMinSet = false;
	    compassZMaxSet = false;
		gravityLimitSet = false;
		XMinReady = false;
		XMaxReady = false;
		YMinReady = false;
		YMaxReady = false;
		ZMinReady = false;
		ZMaxReady = false;
		redTimeStamp = -1L;
		whiteTimeStamp = -1L;
		lastCalibratingCompassTimeStamp = 0L;
		setState( ENGINESTATES_CALIBRATING );
	}

	private void setState( int newState ) {
		Log.d( LOG_TAG, "Transitioning from "+getStateName( state )+
				" to "+getStateName( newState )+
				" at sample counter "+sampleCounter );
		if( state != newState ) {
			state = newState;
			if( iStabilizedCompass != null ) {
				try {
					iStabilizedCompass.statusMessage( state );
				} catch( DeadObjectException ex ) {
					Log.e( LOG_TAG,"step() callback", ex );
				} catch( RemoteException ex ) {
					Log.e( LOG_TAG, "RemoteException",ex );
				}
			} else
				Log.d( LOG_TAG, "setState: cannot call back activity");
		}
	}

	private void setStatus( int status ) {
		Log.d( LOG_TAG, "Setting status "+status+
				" at sample counter "+sampleCounter );
		if( iStabilizedCompass != null ) {
			try {
				iStabilizedCompass.statusMessage( status );
			} catch( DeadObjectException ex ) {
				Log.e( LOG_TAG,"step() callback", ex );
			} catch( RemoteException ex ) {
				Log.e( LOG_TAG, "RemoteException",ex );
			}
		} else
			Log.d( LOG_TAG, "setStatus: cannot call back activity");
	}
	
	
	private String getStateName( int state ) {
		String stateName = null;
		switch( state ) {
		case ENGINESTATES_IDLE:
			stateName = "Idle";
			break;
		
		case ENGINESTATES_CALIBRATING:
			stateName = "Calibrating";
			break;
			
		case ENGINESTATES_MEASURING:
			stateName = "Measuring";
			break;
			
		default:
			stateName = "N/A";
			break;
		}
		return stateName;
	}
	
	private void updateSampleCounter() {		
		++sampleCounter;
		if( ( sampleCounter % SAMPLECTR_MOD ) == 0 ) {
			if( iStabilizedCompass == null )
				Log.d( 
					LOG_TAG, 
					"updateSampleCounter() callback: cannot call back (sampleCounter: "+
					sampleCounter+
					")" );
			else {
				Log.d( 
					LOG_TAG, 
					"updateSampleCounter() callback: sampleCounter: "+
					sampleCounter );
				try {
					iStabilizedCompass.sampleCounter( sampleCounter );
				} catch( DeadObjectException ex ) {
					Log.e( LOG_TAG,"step() callback", ex );
				} catch( RemoteException ex ) {
					Log.e( LOG_TAG, "RemoteException",ex );
				}
			}
		}
	}

	private void processSample( SensorEvent sensorEvent ) {
		float values[] = sensorEvent.values;
		if( values.length < 3 )
				return;
		String sensorName = "n/a";
		int sensorType = SENSORTYPE_NA;
		if( sensorEvent.sensor == accelSensor ) {
			sensorName="accel";
			sensorType = SENSORTYPE_ACCEL;
		} else
		if( sensorEvent.sensor == compassSensor ) {
			sensorName = "compass";
			sensorType = SENSORTYPE_COMPASS;
		}
		if( captureFile != null ) {
			captureFile.println( sensorEvent.timestamp+
					","+
					sensorName+
					","+
					values[0]+
					","+
					values[1]+
					","+
					values[2]);
		}
		updateSampleCounter();
		switch( state ) {
		case ENGINESTATES_CALIBRATING:
			processCalibrating( sensorEvent.timestamp, sensorType, values );
			break;
			
		case ENGINESTATES_MEASURING:
			processMeasuring( sensorEvent.timestamp, sensorType, values );
			break;
		}
	}
	
	private void processCalibrating( long timeStamp, int sensorType, float values [] ) {
		double dValues[] = new double[3];
		dValues[IDX_X] = (double)values[IDX_X];
		dValues[IDX_Y] = (double)values[IDX_Y];
		dValues[IDX_Z] = (double)values[IDX_Z];
		if( sensorType == SENSORTYPE_ACCEL ) {
			if( !gravityLimitSet ) {
				gravityRef = vectorLength( dValues );
				gravityLimitSet = true;
				calibrationGravityLowLimit = -( gravityRef * ( 1.0 - CALIBRATING_ACCEL_LIMIT ) );
				calibrationGravityHighLimit = gravityRef * ( 1.0 - CALIBRATING_ACCEL_LIMIT ); 
				Log.d(LOG_TAG, 
						"reference gravity: "+gravityRef+
						"; gravityLowLimit: "+calibrationGravityLowLimit+
						"; gravityHighLimit: "+calibrationGravityHighLimit);
			}
			if( gravityLimitSet && 
				( ( timeStamp - lastCalibratingCompassTimeStamp ) < MAX_ACCEL_COMPASS_DIFF ) ) {
				boolean XReadyTransition = false;
				boolean YReadyTransition = false;
				boolean ZReadyTransition = false;
// X max/min
				if( dValues[IDX_X] < calibrationGravityLowLimit ) {
					if( !XMinReady && XMaxReady )
						XReadyTransition = true;
					XMinReady = true;
				}
				if( dValues[IDX_X] > calibrationGravityHighLimit ) {
					if( !XMaxReady && XMinReady )
						XReadyTransition = true;
					XMaxReady = true;
				}
// Y max/min
				if( dValues[IDX_Y] < calibrationGravityLowLimit ) {
					if( !YMinReady && YMaxReady )
						YReadyTransition = true;
					YMinReady = true;
				}
				if( dValues[IDX_Y] > calibrationGravityHighLimit ) {
					if( !YMaxReady && YMinReady )
						YReadyTransition = true;
					YMaxReady = true;
				}
// Z max/min
				if( dValues[IDX_Z] < calibrationGravityLowLimit ) {
					if( !ZMinReady && ZMaxReady )
						ZReadyTransition = true;
					ZMinReady = true;
				}
				if( dValues[IDX_Z] > calibrationGravityHighLimit ) {
					if( !ZMaxReady && ZMinReady )
						ZReadyTransition = true;
					ZMaxReady = true;
				}
				if( XReadyTransition )
					setStatus( ENGINESTATES_XGONE );
				if( YReadyTransition )
					setStatus( ENGINESTATES_YGONE );
				if( ZReadyTransition )
					setStatus( ENGINESTATES_ZGONE );
				if( XMinReady && 
					XMaxReady &&
					YMinReady &&
					YMaxReady &&
					ZMinReady &&
					ZMaxReady ) {
// Calibration finished
					compassOffset = new double[3];
					compassOffset[IDX_X] = ( compassXMin + compassXMax ) / 2;
					compassOffset[IDX_Y]= ( compassYMin + compassYMax ) / 2;
					compassOffset[IDX_Z] = ( compassZMin + compassZMax ) / 2;
					Log.d( LOG_TAG, "Offsets; x: "+compassOffset[IDX_X]+
							"; y: "+compassOffset[IDX_Y]+
							"; z: "+compassOffset[IDX_Z]);
					gravityNoMotionLowLimit = gravityRef * ( 1.0 - NOMOTION_ACCEL_LIMIT );
					gravityNoMotionHighLimit = gravityRef * ( 1.0 + NOMOTION_ACCEL_LIMIT );
					lastRefAccel = null;
					lastRefAccelTimeStamp = 0L;
					setState( ENGINESTATES_MEASURING );
				}
			}
		} else
		if( sensorType == SENSORTYPE_COMPASS ) {
			lastCalibratingCompassTimeStamp = timeStamp;
			lastCalibratingCompass = dValues;
			if( !compassXMinSet || ( dValues[IDX_X] < compassXMin ) ) {
				compassXMinSet = true;
				compassXMin = dValues[IDX_X];
			}
			if( !compassXMaxSet || ( dValues[IDX_X] > compassXMax ) ) {
				compassXMaxSet = true;
				compassXMax = dValues[IDX_X];
			}
			if( !compassYMinSet || ( dValues[IDX_Y] < compassYMin ) ) {
				compassYMinSet = true;
				compassYMin = dValues[IDX_Y];
			}
			if( !compassYMaxSet || ( dValues[IDX_Y] > compassYMax ) ) {
				compassYMaxSet = true;
				compassYMax = dValues[IDX_Y];
			}
			if( !compassZMinSet || ( dValues[IDX_Z] < compassZMin ) ) {
				compassZMinSet = true;
				compassZMin = dValues[IDX_Z];
			}
			if( !compassZMaxSet || ( dValues[IDX_Z] > compassZMax ) ) {
				compassZMaxSet = true;
				compassZMax = dValues[IDX_Z];
			}
		}
	}

	private void processMeasuring( long timeStamp, int sensorType, float values [] ) {
		double dValues[] = new double[3];
		dValues[IDX_X] = (double)values[IDX_X];
		dValues[IDX_Y] = (double)values[IDX_Y];
		dValues[IDX_Z] = (double)values[IDX_Z];
		if( sensorType == SENSORTYPE_ACCEL ) {
			double accelLength = vectorLength( dValues );
			if( ( accelLength > gravityNoMotionLowLimit ) && 
				( accelLength < gravityNoMotionHighLimit) ) {
// No motion acceleration - save the acceleration vector
				lastRefAccel = dValues;
				lastRefAccelTimeStamp = timeStamp;
			}
		} else
		if( sensorType == SENSORTYPE_COMPASS ) {
			double correctedCompass[] = vectorSub( dValues, compassOffset );
			double degree = Math.atan2( correctedCompass[IDX_Y], correctedCompass[IDX_X]);
			if( captureFile != null ) {
				captureFile.println( timeStamp+
						","+
						"correctedcompass"+
						","+
						correctedCompass[0]+
						","+
						correctedCompass[1]+
						","+
						correctedCompass[2]);
				captureFile.println( timeStamp+","+
							"degree"+
							","+
							degree);
			}
			sendArrow( true, degree );
// Do we have a usable acceleration vector?
			long diff = timeStamp - lastRefAccelTimeStamp; 
			if( diff < MAX_ACCEL_COMPASS_DIFF ) {
				double rotatedCompass[] = rotateByAccel( correctedCompass,lastRefAccel );
				degree = Math.atan2( rotatedCompass[IDX_Y], rotatedCompass[IDX_X]);
				sendArrow( false, degree );
			}
		}
	}		
	
	private double vectorLength( double vec[] ) {
		return Math.sqrt( ( vec[IDX_X]*vec[IDX_X] )+
						  ( vec[IDX_Y]*vec[IDX_Y] )+
						  ( vec[IDX_Z]*vec[IDX_Z] ) );
	}

	private double[] vectorSub( double v1[],double v2[]) {
		double r[] = new double[3];
		r[IDX_X] = v1[IDX_X]-v2[IDX_X];
		r[IDX_Y] = v1[IDX_Y]-v2[IDX_Y];
		r[IDX_Z] = v1[IDX_Z]-v2[IDX_Z];
		return r;
	}
	
	private double[] vectorScalarMul( double vec[], double scalar ) {
		double r[] = new double[3];
		r[IDX_X] = vec[IDX_X]*scalar;
		r[IDX_Y] = vec[IDX_Y]*scalar;
		r[IDX_Z] = vec[IDX_Z]*scalar;
		return r;
	}
	
	private double vectorAngle( double v1[],double v2[]) {
		double ab=vectorLength(v1)*vectorLength(v2);
		double cosfi=( (v1[IDX_X]*v2[IDX_X])+
					   (v1[IDX_Y]*v2[IDX_Y])+
					   (v1[IDX_Z]*v2[IDX_Z])) / ab;
		if( cosfi>=1.0 )		// Shouldn't happen but rounding errors cause it sometimes
					cosfi=1.0;
		if( cosfi<=-1.0 )
					cosfi=-1.0;
		return Math.acos(cosfi);		
	}
	
	private double[] vectorCrossProductU( double a[], double b[]) {
		double cp[] = new double[3];
		cp[IDX_X]= a[IDX_Y]*b[IDX_Z]-a[IDX_Z]*b[IDX_Y];
		cp[IDX_Y]= a[IDX_Z]*b[IDX_X]-a[IDX_X]*b[IDX_Z];
		cp[IDX_Z]= a[IDX_X]*b[IDX_Y]-a[IDX_Y]*b[IDX_X];
		double amp = vectorLength( cp );
		if( amp==0.0 ) {
			cp[IDX_X] = 0.0;
			cp[IDX_Y] = 0.0;
			cp[IDX_Z] = 0.0;
		} else {
			cp[IDX_X] /= amp;
			cp[IDX_Y] /= amp;
			cp[IDX_Z] /= amp;
		}
		return cp;		
	}
	
	private double[] matrixMul( double r[][], double vec[] ) {
		double rv[] = new double[3];
		rv[IDX_X]=r[0][0]*vec[IDX_X]+r[0][1]*vec[IDX_Y]+r[0][2]*vec[IDX_Z];
		rv[IDX_Y]=r[1][0]*vec[IDX_X]+r[1][1]*vec[IDX_Y]+r[1][2]*vec[IDX_Z];
		rv[IDX_Z]=r[2][0]*vec[IDX_X]+r[2][1]*vec[IDX_Y]+r[2][2]*vec[IDX_Z];
		return rv;		
	}
	
	private double[] rotateAroundAxis( double vec[], double u[], double angle) {
		double ca = Math.cos( angle );
		double sa = Math.sin( angle );
		double omca = 1 - ca;
		double r[][] = new double[3][3];
		r[0][0]=ca+(u[0]*u[0]*omca);
		r[0][1]=(u[0]*u[1]*omca)-u[2]*sa;
		r[0][2]=(u[0]*u[2]*omca)+u[1]*sa;
		r[1][0]=(u[1]*u[0]*omca)+u[2]*sa;
		r[1][1]=ca+(u[1]*u[1]*omca);
		r[1][2]=(u[1]*u[2]*omca)-u[0]*sa;
		r[2][0]=(u[2]*u[0]*omca)-u[1]*sa;
		r[2][1]=(u[2]*u[1]*omca)+u[0]*sa;
		r[2][2]=ca+(u[2]*u[2]*omca);
		double rv[]=matrixMul(r,vec);
		return rv;
	}
		
	private Object[] calculateRotationVector( double v1[], double v2[]) {
		double angle=vectorAngle(v1,v2 );
		double u[] = vectorCrossProductU( v1,v2);
		Object r[] = new Object[2];
		r[0] = u;
		r[1] = new Double( angle );
		return r;
	}
		
	private double[] rotateByAccel( double compassVector[], double accelVector[] ) {
		double zaxis[] = { 0.0, 0.0, 1.0 };
		Object r[] = calculateRotationVector( accelVector, zaxis );
		double u[] = (double[])r[0];
		Double angle = (Double)r[1];
		double rotatedCompass[] = rotateAroundAxis( compassVector, u, angle.doubleValue());
		return rotatedCompass;
	}

	private void sendArrow( boolean isRedArrow, double degree ) {
		long currentTime = System.currentTimeMillis();
		long prevTimeStamp = isRedArrow ? redTimeStamp : whiteTimeStamp;
		long tdiff =  currentTime - prevTimeStamp;
		if( ( prevTimeStamp < 0L ) || ( tdiff > DIFF_UPDATE_TIMEOUT )) {
			Log.d( LOG_TAG, "sendArrow; isRedArrow: "+isRedArrow+"; degree: "+degree );
			if( iStabilizedCompass != null ) {
				try {
					if( isRedArrow )
						iStabilizedCompass.redArrow( degree );
					else
						iStabilizedCompass.whiteArrow(degree);
				} catch( DeadObjectException ex ) {
					Log.e( LOG_TAG,"step() callback", ex );
				} catch( RemoteException ex ) {
					Log.e( LOG_TAG, "RemoteException",ex );
				}
				if( isRedArrow )
					redTimeStamp = currentTime;
				else
					whiteTimeStamp = currentTime;
			} else
					Log.d(LOG_TAG, "sendArrow: cannot call back main activity");
		}
	}
	
    private IStabilizedCompass iStabilizedCompass = null;
    private final ISamplingService.Stub serviceBinder = 
			new ISamplingService.Stub() {
		public void setCallback( IBinder binder ) {
			Log.d( LOG_TAG,"setCallback" );
			iStabilizedCompass = IStabilizedCompass.Stub.asInterface( binder );
      	}

		public void removeCallback() {
			Log.d( LOG_TAG, "removeCallback" );
			iStabilizedCompass = null;
		}

		public boolean isSampling() {
			return samplingStarted;
		}

		public void stopSampling() {
			Log.d( LOG_TAG, "stopSampling" );
			SamplingService.this.stopSampling();
			stopSelf();
		}

		@Override
		public int getState() throws RemoteException {
			return state;
		}

    };

    
    private boolean XMinReady;
    private boolean XMaxReady;
    private boolean YMinReady;
    private boolean YMaxReady;
    private boolean ZMinReady;
    private boolean ZMaxReady;
    private boolean compassXMinSet;
    private double compassXMin;
    private boolean compassXMaxSet;
    private double compassXMax;
    private boolean compassYMinSet;
    private double compassYMin;
    private boolean compassYMaxSet;
    private double compassYMax;
    private boolean compassZMinSet;
    private double compassZMin;
    private boolean compassZMaxSet;
    private double compassZMax;
    private boolean gravityLimitSet;
    private double gravityRef;
    private double calibrationGravityLowLimit;
    private double calibrationGravityHighLimit;
    private double gravityNoMotionLowLimit;
    private double gravityNoMotionHighLimit;
    private double lastRefAccel[];
    private long lastRefAccelTimeStamp;
    private double compassOffset[];
    private long lastCalibratingCompassTimeStamp;
    private double lastCalibratingCompass[];
    private int sampleCounter;
    private long redTimeStamp;
    private long whiteTimeStamp;
    private int rate;
    private SensorManager sensorManager;
    private PrintWriter captureFile = null;
	private Sensor accelSensor;
	private Sensor compassSensor;
	private boolean samplingStarted = false;
	private int state;
}

